@@ -29,11 +29,20 @@
 #include <assert.h>
 #include <stdarg.h>
 #include <ctype.h>
+#if HAVE_SYS_TYPES_H
+# include <sys/types.h>
+#endif
 
-/* By default, assume we have strtoul. */
+/* By default, assume we have inttypes.h, strtoul, and uintptr_t. */
 #if !defined(HAVE_STRTOUL) && !defined(HAVE_CONFIG_H)
 # define HAVE_STRTOUL 1
 #endif
+#if defined(HAVE_INTTYPES_H) || !defined(HAVE_CONFIG_H)
+# include <inttypes.h>
+#endif
+#if !defined(HAVE_UINTPTR_T) && defined(HAVE_CONFIG_H)
+typedef unsigned long uintptr_t;
+#endif
 
 #ifdef __cplusplus
 extern "C" {
@@ -226,17 +235,24 @@ struct Clp_ParserState {
 
 
 typedef struct Clp_StringList {
-
     Clp_Option *items;
     Clp_InternOption *iopt;
     int nitems;
 
-    int allow_int;
+    unsigned char allow_int;
+    unsigned char val_long;
     int nitems_invalid_report;
-
 } Clp_StringList;
 
 
+static const Clp_Option clp_option_sentinel[] = {
+    {"", 0, Clp_NotOption, 0, 0},
+    {"", 0, Clp_Done, 0, 0},
+    {"", 0, Clp_BadOption, 0, 0},
+    {"", 0, Clp_Error, 0, 0}
+};
+
+
 static int parse_string(Clp_Parser *, const char *, int, void *);
 static int parse_int(Clp_Parser *, const char *, int, void *);
 static int parse_bool(Clp_Parser *, const char *, int, void *);
@@ -434,6 +450,18 @@ compare_options(Clp_Parser *clp, const Clp_Option *o1, Clp_InternOption *io1,
     }
 }
 
+static void
+calculate_lmm(Clp_Parser *clp, const Clp_Option *opt, Clp_InternOption *iopt, int nopt)
+{
+    int i, j;
+    for (i = 0; i < nopt; ++i) {
+	iopt[i].lmmpos = iopt[i].lmmneg = 1;
+	iopt[i].lmmpos_short = iopt[i].lmmneg_short = 0;
+	for (j = 0; j < nopt; ++j)
+	    compare_options(clp, &opt[i], &iopt[i], &opt[j], &iopt[j]);
+    }
+}
+
 /** @param argc number of arguments
  * @param argv argument array
  * @param nopt number of option definitions
@@ -455,7 +483,8 @@ compare_options(Clp_Parser *clp, const Clp_Option *o1, Clp_InternOption *io1,
  * one of the substrings "UTF-8", "UTF8", or "utf8".  Override this with
  * Clp_SetUTF8().</li>
  * <li>The Clp_ValString, Clp_ValStringNotOption, Clp_ValInt, Clp_ValUnsigned,
- * Clp_ValBool, and Clp_ValDouble types are installed.</li>
+ * Clp_ValLong, Clp_ValUnsignedLong, Clp_ValBool, and Clp_ValDouble types are
+ * installed.</li>
  * <li>Errors are reported to standard error.</li>
  * </ul>
  *
@@ -480,7 +509,7 @@ Clp_NewParser(int argc, const char * const *argv, int nopt, const Clp_Option *op
     if (!clp || !cli || !iopt || !cli->valtype)
 	goto failed;
 
-    clp->opt = -1;
+    clp->option = &clp_option_sentinel[-Clp_Done];
     clp->negated = 0;
     clp->have_val = 0;
     clp->vstr = 0;
@@ -518,8 +547,10 @@ Clp_NewParser(int argc, const char * const *argv, int nopt, const Clp_Option *op
     cli->nvaltype = 0;
     Clp_AddType(clp, Clp_ValString, 0, parse_string, 0);
     Clp_AddType(clp, Clp_ValStringNotOption, Clp_DisallowOptions, parse_string, 0);
-    Clp_AddType(clp, Clp_ValInt, 0, parse_int, 0);
-    Clp_AddType(clp, Clp_ValUnsigned, 0, parse_int, (void *)cli);
+    Clp_AddType(clp, Clp_ValInt, 0, parse_int, (void*) (uintptr_t) 0);
+    Clp_AddType(clp, Clp_ValUnsigned, 0, parse_int, (void*) (uintptr_t) 1);
+    Clp_AddType(clp, Clp_ValLong, 0, parse_int, (void*) (uintptr_t) 2);
+    Clp_AddType(clp, Clp_ValUnsignedLong, 0, parse_int, (void*) (uintptr_t) 3);
     Clp_AddType(clp, Clp_ValBool, 0, parse_bool, 0);
     Clp_AddType(clp, Clp_ValDouble, 0, parse_double, 0);
 
@@ -603,15 +634,9 @@ int
 Clp_SetUTF8(Clp_Parser *clp, int utf8)
 {
     Clp_Internal *cli = clp->internal;
-    int i, j, old_utf8 = cli->utf8;
+    int old_utf8 = cli->utf8;
     cli->utf8 = utf8;
-    for (i = 0; i < cli->nopt; ++i) {
-	cli->iopt[i].lmmpos = cli->iopt[i].lmmneg = 1;
-	cli->iopt[i].lmmpos_short = cli->iopt[i].lmmneg_short = 0;
-	for (j = 0; j < cli->nopt; ++j)
-	    compare_options(clp, &cli->opt[i], &cli->iopt[i],
-			    &cli->opt[j], &cli->iopt[j]);
-    }
+    calculate_lmm(clp, cli->opt, cli->iopt, cli->nopt);
     return old_utf8;
 }
 
@@ -683,7 +708,7 @@ Clp_OptionChar(Clp_Parser *clp, int c)
 int
 Clp_SetOptionChar(Clp_Parser *clp, int c, int type)
 {
-    int i, j, long1pos, long1neg;
+    int i, long1pos, long1neg;
     int old = Clp_OptionChar(clp, c);
     Clp_Internal *cli = clp->internal;
 
@@ -722,13 +747,7 @@ Clp_SetOptionChar(Clp_Parser *clp, int c, int type)
 	/* Must recheck option set */
 	cli->long1pos = long1pos;
 	cli->long1neg = long1neg;
-	for (i = 0; i < cli->nopt; ++i) {
-	    cli->iopt[i].lmmpos = cli->iopt[i].lmmneg = 1;
-	    cli->iopt[i].lmmpos_short = cli->iopt[i].lmmneg_short = 0;
-	    for (j = 0; j < cli->nopt; ++j)
-		compare_options(clp, &cli->opt[i], &cli->iopt[i],
-				&cli->opt[j], &cli->iopt[j]);
-	}
+	calculate_lmm(clp, cli->opt, cli->iopt, cli->nopt);
     }
 
     return old;
@@ -766,7 +785,7 @@ Clp_SetOptions(Clp_Parser *clp, int nopt, const Clp_Option *opt)
 {
     Clp_Internal *cli = clp->internal;
     Clp_InternOption *iopt;
-    int i, j;
+    int i;
     static unsigned opt_generation = 0;
 
     if (nopt > cli->nopt) {
@@ -823,12 +842,7 @@ Clp_SetOptions(Clp_Parser *clp, int nopt, const Clp_Option *opt)
     }
 
     /* Check option set */
-    for (i = 0; i < nopt; ++i) {
-	iopt[i].lmmpos = iopt[i].lmmneg = 1;
-	iopt[i].lmmpos_short = iopt[i].lmmneg_short = 0;
-	for (j = 0; j < nopt; ++j)
-	    compare_options(clp, &opt[i], &iopt[i], &opt[j], &iopt[j]);
-    }
+    calculate_lmm(clp, opt, iopt, nopt);
 
     return 0;
 }
@@ -1076,28 +1090,31 @@ parse_string(Clp_Parser *clp, const char *arg, int complain, void *user_data)
 }
 
 static int
-parse_int(Clp_Parser *clp, const char *arg, int complain, void *user_data)
+parse_int(Clp_Parser *clp, const char *arg, int complain, void* user_data)
 {
     const char *val;
+    uintptr_t type = (uintptr_t) user_data;
     if (*arg == 0 || isspace((unsigned char) *arg)
-	|| (user_data != 0 && *arg == '-'))
+	|| ((type & 1) && *arg == '-'))
 	val = arg;
-    else if (user_data != 0) {	/* unsigned */
+    else if (type & 1) { // unsigned
 #if HAVE_STRTOUL
-	clp->val.u = strtoul(arg, (char **) &val, 0);
+	clp->val.ul = strtoul(arg, (char **) &val, 0);
 #else
 	/* don't bother really trying to do it right */
 	if (arg[0] == '-')
 	    val = arg;
 	else
-	    clp->val.u = strtol(arg, (char **) &val, 0);
+	    clp->val.l = strtol(arg, (char **) &val, 0);
 #endif
     } else
-	clp->val.i = strtol(arg, (char **) &val, 0);
+	clp->val.l = strtol(arg, (char **) &val, 0);
+    if (type <= 1)
+        clp->val.u = (unsigned) clp->val.ul;
     if (*arg != 0 && *val == 0)
 	return 1;
     else if (complain) {
-	const char *message = user_data != 0
+	const char *message = type & 1
 	    ? "%<%O%> expects a nonnegative integer, not %<%s%>"
 	    : "%<%O%> expects an integer, not %<%s%>";
 	return Clp_OptionError(clp, message, arg);
@@ -1171,27 +1188,26 @@ parse_string_list(Clp_Parser *clp, const char *arg, int complain, void *user_dat
 	 &ambiguous, ambiguous_values);
     if (idx >= 0) {
 	clp->val.i = sl->items[idx].option_id;
-	return 1;
+        if (sl->val_long)
+            clp->val.l = clp->val.i;
+        return 1;
     }
 
     if (sl->allow_int) {
-	if (parse_int(clp, arg, 0, 0))
+	if (parse_int(clp, arg, 0, (void*) (uintptr_t) (sl->val_long ? 2 : 0)))
 	    return 1;
     }
 
     if (complain) {
-	const char *complaint;
-	if (ambiguous)
-	    complaint = "ambiguous value %<%s%> for option %<%O%>";
-	else {
-	    complaint = "unknown value %<%s%> for option %<%O%>";
+	const char *complaint = (ambiguous ? "ambiguous" : "invalid");
+	if (!ambiguous) {
 	    ambiguous = sl->nitems_invalid_report;
 	    for (idx = 0; idx < ambiguous; idx++)
 		ambiguous_values[idx] = idx;
 	}
 	return ambiguity_error
 	    (clp, ambiguous, ambiguous_values, sl->items, sl->iopt,
-	     "", complaint, arg);
+	     "", "option %<%O%> value %<%s%> is %s", arg, complaint);
     } else
 	return 0;
 }
@@ -1201,7 +1217,7 @@ static int
 finish_string_list(Clp_Parser *clp, int val_type, int flags,
 		   Clp_Option *items, int nitems, int itemscap)
 {
-    int i, j;
+    int i;
     Clp_StringList *clsl = (Clp_StringList *)malloc(sizeof(Clp_StringList));
     Clp_InternOption *iopt = (Clp_InternOption *)malloc(sizeof(Clp_InternOption) * nitems);
     if (!clsl || !iopt)
@@ -1211,6 +1227,7 @@ finish_string_list(Clp_Parser *clp, int val_type, int flags,
     clsl->iopt = iopt;
     clsl->nitems = nitems;
     clsl->allow_int = (flags & Clp_AllowNumbers) != 0;
+    clsl->val_long = (flags & Clp_StringListLong) != 0;
 
     if (nitems < MAX_AMBIGUOUS_VALUES && nitems < itemscap && clsl->allow_int) {
 	items[nitems].long_name = "any integer";
@@ -1223,12 +1240,8 @@ finish_string_list(Clp_Parser *clp, int val_type, int flags,
     for (i = 0; i < nitems; i++) {
 	iopt[i].ilong = iopt[i].ipos = 1;
 	iopt[i].ishort = iopt[i].ineg = iopt[i].ilongoff = iopt[i].iprefmatch = 0;
-	iopt[i].lmmpos = 1;
-	iopt[i].lmmpos_short = 0;
     }
-    for (i = 0; i < nitems; i++)
-	for (j = 0; j < nitems; j++)
-	    compare_options(clp, &items[i], &iopt[i], &items[j], &iopt[j]);
+    calculate_lmm(clp, items, iopt, nitems);
 
     if (Clp_AddType(clp, val_type, 0, parse_string_list, clsl) >= 0)
 	return 0;
@@ -1285,7 +1298,12 @@ Clp_AddStringListType(Clp_Parser *clp, int val_type, int flags, ...)
 	char *name = va_arg(val, char *);
 	if (!name)
 	    break;
-	value = va_arg(val, int);
+        if (flags & Clp_StringListLong) {
+            long lvalue = va_arg(val, long);
+            value = (int) lvalue;
+            assert(value == lvalue);
+        } else
+            value = va_arg(val, int);
 
 	if (nitems >= itemscap) {
 	    Clp_Option *new_items;
@@ -1763,8 +1781,10 @@ find_short(Clp_Parser *clp, const char *text)
 
     for (i = 0; i < cli->nopt; i++)
 	if (iopt[i].ishort && opt[i].short_name == c
-	    && (clp->negated ? iopt[i].ineg : iopt[i].ipos))
+            && (!clp->negated || iopt[i].ineg)) {
+            clp->negated = clp->negated || !iopt[i].ipos;
 	    return i;
+        }
 
     return -1;
 }
@@ -1828,8 +1848,9 @@ Clp_Next(Clp_Parser *clp)
     /* Get the next argument or option */
     if (!next_argument(clp, cli->option_processing ? 0 : 2)) {
 	clp->val.s = clp->vstr;
-	clp->opt = clp->have_val ? Clp_NotOption : Clp_Done;
-	return clp->opt;
+	optno = clp->have_val ? Clp_NotOption : Clp_Done;
+	clp->option = &clp_option_sentinel[-optno];
+	return optno;
     }
 
     clp->negated = cli->whole_negated;
@@ -1867,7 +1888,8 @@ Clp_Next(Clp_Parser *clp)
 	    Clp_OptionError(clp, "unrecognized option %<%s%s%>",
 			    cli->option_chars, cli->xtext);
 
-	return (clp->opt = Clp_BadOption);
+	clp->option = &clp_option_sentinel[-Clp_BadOption];
+	return Clp_BadOption;
     }
 
     /* Set the current option */
@@ -1880,20 +1902,27 @@ Clp_Next(Clp_Parser *clp)
 	|| (!cli->iopt[optno].imandatory && !cli->iopt[optno].ioptional)) {
 	if (clp->have_val) {
 	    Clp_OptionError(clp, "%<%O%> can%,t take an argument");
-	    clp->opt = Clp_BadOption;
-	} else
-	    clp->opt = cli->opt[optno].option_id;
-	return clp->opt;
+	    clp->option = &clp_option_sentinel[-Clp_BadOption];
+	    return Clp_BadOption;
+	} else {
+	    clp->option = &cli->opt[optno];
+	    return cli->opt[optno].option_id;
+	}
     }
 
     /* Get an argument if we need one, or if it's optional */
     /* Sanity-check the argument type. */
     opt = &cli->opt[optno];
-    if (opt->val_type <= 0)
-	return (clp->opt = Clp_Error);
+    if (opt->val_type <= 0) {
+	clp->option = &clp_option_sentinel[-Clp_Error];
+	return Clp_Error;
+    }
     vtpos = val_type_binsearch(cli, opt->val_type);
-    if (vtpos == cli->nvaltype || cli->valtype[vtpos].val_type != opt->val_type)
-	return (clp->opt = Clp_Error);
+    if (vtpos == cli->nvaltype
+	|| cli->valtype[vtpos].val_type != opt->val_type) {
+	clp->option = &clp_option_sentinel[-Clp_Error];
+	return Clp_Error;
+    }
 
     /* complain == 1 only if the argument was explicitly given,
        or it is mandatory. */
@@ -1913,31 +1942,34 @@ Clp_Next(Clp_Parser *clp)
 		Clp_OptionError(clp, "%<%O%> requires a non-option argument");
 	    else
 		Clp_OptionError(clp, "%<%O%> requires an argument");
-	    return (clp->opt = Clp_BadOption);
+	    clp->option = &clp_option_sentinel[-Clp_BadOption];
+	    return Clp_BadOption;
 	}
 
     } else if (cli->is_short && !clp->have_val
-	       && cli->xtext[clp_utf8_charlen(cli, cli->xtext)]) {
+	       && cli->xtext[clp_utf8_charlen(cli, cli->xtext)])
 	/* The -[option]argument case:
 	   Assume that the rest of the current string is the argument. */
 	next_argument(clp, 1);
-	complain = 1;
-    }
 
     /* Parse the argument */
+    clp->option = opt;
     if (clp->have_val) {
 	Clp_ValType *atr = &cli->valtype[vtpos];
 	if (atr->func(clp, clp->vstr, complain, atr->user_data) <= 0) {
 	    /* parser failed */
 	    clp->have_val = 0;
-	    if (complain)
-		return (clp->opt = Clp_BadOption);
-	    else
+	    if (cli->iopt[optno].imandatory) {
+		clp->option = &clp_option_sentinel[-Clp_BadOption];
+		return Clp_BadOption;
+	    } else {
 		Clp_RestoreParser(clp, &clpsave);
+                clp->option = opt;
+            }
 	}
     }
 
-    return (clp->opt = opt->option_id);
+    return opt->option_id;
 }
 
 
