@@ -223,7 +223,7 @@ name_color(int color, Gif_Colormap *gfcm, char *buf)
     strcpy(buf, "transparent");
   else {
     Gif_Color *c = &gfcm->col[color];
-    sprintf(buf, "#%02X%02X%02X", c->red, c->green, c->blue);
+    sprintf(buf, "#%02X%02X%02X", c->gfc_red, c->gfc_green, c->gfc_blue);
   }
 }
 
