@@ -162,6 +162,8 @@ void verbose_open(char, const char *);
 void verbose_close(char);
 void verbose_endline(void);
 
+const char* debug_color_str(const Gif_Color* gfc);
+
 #define EXIT_OK		0
 #define EXIT_ERR	1
 #define EXIT_USER_ERR	1
@@ -219,8 +221,8 @@ int	crop_image(Gif_Image *, Gt_Crop *, int preserve_total_crop);
 
 void	flip_image(Gif_Image *, int scr_width, int scr_height, int is_vert);
 void	rotate_image(Gif_Image *, int scr_width, int scr_height, int rotation);
-void	scale_image(Gif_Stream *, Gif_Image *, double xfactor, double yfactor);
-void	resize_stream(Gif_Stream *, int new_width, int new_height, int fit);
+void	resize_stream(Gif_Stream *, double new_width, double new_height,
+                      int fit);
 
 /*****
  * quantization
@@ -230,21 +232,18 @@ Gif_Color *histogram(Gif_Stream *, int *);
 #define COLORMAP_DIVERSITY		0
 #define COLORMAP_BLEND_DIVERSITY	1
 #define COLORMAP_MEDIAN_CUT		2
-Gif_Colormap *colormap_blend_diversity(Gif_Color *, int, int);
-Gif_Colormap *colormap_flat_diversity(Gif_Color *, int, int);
-Gif_Colormap *colormap_median_cut(Gif_Color *, int, int);
+Gif_Colormap *colormap_blend_diversity(Gif_Color *, int, int, int);
+Gif_Colormap *colormap_flat_diversity(Gif_Color *, int, int, int);
+Gif_Colormap *colormap_median_cut(Gif_Color *, int, int, int);
 
-typedef struct color_hash_item color_hash_item;
+typedef struct kd3_tree kd3_tree;
 typedef void (*colormap_image_func)
-     (Gif_Image *, uint8_t *, Gif_Colormap *, Gif_Colormap *,
-      color_hash_item **, uint32_t *);
+     (Gif_Image*, uint8_t*, Gif_Colormap*, kd3_tree*, uint32_t*);
 
 void	colormap_image_posterize
-	(Gif_Image *, uint8_t *, Gif_Colormap *, Gif_Colormap *,
-	 color_hash_item **, uint32_t *);
+        (Gif_Image*, uint8_t*, Gif_Colormap*, kd3_tree*, uint32_t*);
 void	colormap_image_floyd_steinberg
-	(Gif_Image *, uint8_t *, Gif_Colormap *, Gif_Colormap *,
-	 color_hash_item **, uint32_t *);
+        (Gif_Image*, uint8_t*, Gif_Colormap*, kd3_tree*, uint32_t*);
 void	colormap_stream(Gif_Stream *, Gif_Colormap *, colormap_image_func);
 
 /*****
