@@ -376,81 +376,85 @@ static int
 expand_difference_bounds(Gif_OptData *bounds, Gif_Image *this_bounds)
 {
   int x, y, expanded = 0;
-
-  int lf = bounds->left, tp = bounds->top,
-      rt = lf + bounds->width - 1, bt = tp + bounds->height - 1;
-
   Gif_OptBounds ob = safe_bounds(this_bounds);
-  int tlf = ob.left, ttp = ob.top,
-      trt = ob.left + ob.width - 1, tbt = ob.top + ob.height - 1;
 
-  if (lf > rt || tp > bt)
-    lf = 0, tp = 0, rt = screen_width - 1, bt = screen_height - 1;
+  if (bounds->width <= 0 || bounds->height <= 0) {
+      bounds->left = bounds->top = 0;
+      bounds->width = screen_width;
+      bounds->height = screen_height;
+  }
 
-  for (y = ttp; y < tp; y++) {
-    uint16_t *now = this_data + screen_width * y;
-    uint16_t *next = next_data + screen_width * y;
-    for (x = tlf; x <= trt; x++)
+  /* 20.Nov.2013 - The image `bounds` might be larger than `this_bounds`
+     because of a previous frame's background disposal. Don't accidentally
+     shrink `this_bounds`. */
+  if (ob.left > bounds->left) {
+      ob.width = (ob.left + ob.width) - bounds->left;
+      ob.left = bounds->left;
+  }
+  if (ob.top > bounds->top) {
+      ob.height = (ob.top + ob.height) - bounds->top;
+      ob.top = bounds->top;
+  }
+  if (ob.left + ob.width < bounds->left + bounds->width)
+      ob.width = bounds->left + bounds->width - ob.left;
+  if (ob.top + ob.height < bounds->top + bounds->height)
+      ob.height = bounds->top + bounds->height - ob.top;
+
+  for (; ob.top < bounds->top; ++ob.top, --ob.height) {
+    uint16_t *now = this_data + screen_width * ob.top;
+    uint16_t *next = next_data + screen_width * ob.top;
+    for (x = ob.left; x < ob.left + ob.width; ++x)
       if (now[x] != TRANSP && next[x] == TRANSP) {
 	expanded = 1;
-	goto found_top;
+	break;
       }
   }
- found_top:
-  tp = y;
 
-  for (y = tbt; y > bt; y--) {
-    uint16_t *now = this_data + screen_width * y;
-    uint16_t *next = next_data + screen_width * y;
-    for (x = tlf; x <= trt; x++)
+  for (; ob.top + ob.height > bounds->top + bounds->height; --ob.height) {
+    uint16_t *now = this_data + screen_width * (ob.top + ob.height - 1);
+    uint16_t *next = next_data + screen_width * (ob.top + ob.height - 1);
+    for (x = ob.left; x < ob.left + ob.width; ++x)
       if (now[x] != TRANSP && next[x] == TRANSP) {
 	expanded = 1;
-	goto found_bottom;
+	break;
       }
   }
- found_bottom:
-  bt = y;
 
-  for (x = tlf; x < lf; x++) {
-    uint16_t *now = this_data + x;
-    uint16_t *next = next_data + x;
-    for (y = tp; y <= bt; y++)
+  for (; ob.left < bounds->left; ++ob.left, --ob.width) {
+    uint16_t *now = this_data + ob.left;
+    uint16_t *next = next_data + ob.left;
+    for (y = ob.top; y < ob.top + ob.height; ++y)
       if (now[y*screen_width] != TRANSP && next[y*screen_width] == TRANSP) {
 	expanded = 1;
-	goto found_left;
+	break;
       }
   }
- found_left:
-  lf = x;
 
-  for (x = trt; x > rt; x--) {
-    uint16_t *now = this_data + x;
-    uint16_t *next = next_data + x;
-    for (y = tp; y <= bt; y++)
+  for (; ob.left + ob.width > bounds->left + bounds->width; --ob.width) {
+    uint16_t *now = this_data + ob.left + ob.width - 1;
+    uint16_t *next = next_data + ob.left + ob.width - 1;
+    for (y = ob.top; y < ob.top + ob.height; ++y)
       if (now[y*screen_width] != TRANSP && next[y*screen_width] == TRANSP) {
 	expanded = 1;
-	goto found_right;
+	break;
       }
   }
- found_right:
-  rt = x;
 
   if (!expanded)
-    for (y = tp; y <= bt; ++y) {
+    for (y = ob.top; y < ob.top + ob.height; ++y) {
       uint16_t *now = this_data + y*screen_width;
       uint16_t *next = next_data + y*screen_width;
-      for (x = lf; x <= rt; ++x)
+      for (x = ob.left; x < ob.left + ob.width; ++x)
 	if (now[x] != TRANSP && next[x] == TRANSP) {
 	  expanded = 1;
-	  goto found_expanded;
+	  break;
 	}
     }
 
- found_expanded:
-  bounds->left = lf;
-  bounds->top = tp;
-  bounds->width = rt + 1 - lf;
-  bounds->height = bt + 1 - tp;
+  bounds->left = ob.left;
+  bounds->top = ob.top;
+  bounds->width = ob.width;
+  bounds->height = ob.height;
   return expanded;
 }
 
@@ -468,8 +472,8 @@ fix_difference_bounds(Gif_OptData *bounds)
   }
   /* assert that image lies completely within screen */
   assert(bounds->top < screen_height && bounds->left < screen_width
-	 && bounds->top + bounds->height - 1 < screen_height
-	 && bounds->left + bounds->width - 1 < screen_width);
+	 && bounds->top + bounds->height <= screen_height
+	 && bounds->left + bounds->width <= screen_width);
 }
 
 
@@ -545,9 +549,6 @@ get_used_colors(Gif_OptData *bounds, int use_transparency)
 	  need[i] = REQUIRED;
       count[REQUIRED] += count[REPLACE_TRANSP];
     }
-    /* If too many "actually used" pixels, fail miserably */
-    if (count[REQUIRED] > 256)
-      fatal_error("more than 256 colors required in a frame", count[REQUIRED]);
     /* If we can afford to have transparency, and we want to use it, then
        include it */
     if (count[REQUIRED] < 256 && use_transparency && !need[TRANSP]) {
@@ -572,6 +573,7 @@ create_subimages(Gif_Stream *gfs, int optimize_flags, int save_uncompressed)
   Gif_Image *last_gfi;
   int next_data_valid;
   uint16_t *previous_data = 0;
+  int local_color_tables = 0;
 
   screen_size = screen_width * screen_height;
 
@@ -583,15 +585,22 @@ create_subimages(Gif_Stream *gfs, int optimize_flags, int save_uncompressed)
   erase_screen(this_data);
   last_gfi = 0;
 
-  /* PRECONDITION: last_data, previous_data -- garbage
-     this_data -- equal to image data after disposal of previous image
-     next_data -- equal to image data for next image if next_image_valid */
+  /* PRECONDITION:
+     previous_data -- garbage
+     last_data -- optimized image after disposal of previous optimized frame
+     this_data -- input image after disposal of previous input frame
+     next_data -- input image after application of current input frame,
+                  if next_image_valid */
   for (image_index = 0; image_index < gfs->nimages; image_index++) {
     Gif_Image *gfi = gfs->images[image_index];
     Gif_OptData *subimage = new_opt_data();
+    if (gfi->local)
+        local_color_tables = 1;
 
     /* save previous data if necessary */
-    if (gfi->disposal == GIF_DISPOSAL_PREVIOUS) {
+    if (gfi->disposal == GIF_DISPOSAL_PREVIOUS
+        || (local_color_tables && image_index > 0
+            && last_gfi->disposal > GIF_DISPOSAL_ASIS)) {
       if (!previous_data)
 	previous_data = Gif_NewArray(uint16_t, screen_size);
       memcpy(previous_data, this_data, sizeof(uint16_t) * screen_size);
@@ -606,6 +615,7 @@ create_subimages(Gif_Stream *gfs, int optimize_flags, int save_uncompressed)
     } else
       apply_frame(this_data, gfi, 0, save_uncompressed);
 
+ retry_frame:
     /* find minimum area of difference between this image and last image */
     subimage->disposal = GIF_DISPOSAL_ASIS;
     if (image_index > 0)
@@ -638,10 +648,29 @@ create_subimages(Gif_Stream *gfs, int optimize_flags, int save_uncompressed)
 
     /* set map of used colors */
     {
-      int use_transparency = (optimize_flags & GT_OPT_MASK) > 1 && image_index > 0;
+      int use_transparency = (optimize_flags & GT_OPT_MASK) > 1
+          && image_index > 0;
       if (image_index == 0 && background == TRANSP)
 	use_transparency = 2;
       get_used_colors(subimage, use_transparency);
+      /* Gifsicle's optimization strategy normally creates frames with ASIS
+         or BACKGROUND disposal (not PREVIOUS disposal). However, there are
+         cases when PREVIOUS disposal is strictly required, or a frame would
+         require more than 256 colors. Detect this case and try to recover. */
+      if (subimage->required_color_count > 256) {
+          if (image_index > 0 && local_color_tables) {
+              Gif_OptData *subimage = (Gif_OptData*) last_gfi->user_data;
+              if ((last_gfi->disposal == GIF_DISPOSAL_PREVIOUS
+                   || last_gfi->disposal == GIF_DISPOSAL_BACKGROUND)
+                  && subimage->disposal != last_gfi->disposal) {
+                  subimage->disposal = last_gfi->disposal;
+                  memcpy(last_data, previous_data, sizeof(uint16_t) * screen_size);
+                  goto retry_frame;
+              }
+          }
+          fatal_error("%d colors required in a frame (256 is max)",
+                      subimage->required_color_count);
+      }
     }
 
     gfi->user_data = subimage;
@@ -815,8 +844,8 @@ colormap_rgb_permutation_sorter(const void *v1, const void *v2)
 {
   const Gif_Color *col1 = (const Gif_Color *)v1;
   const Gif_Color *col2 = (const Gif_Color *)v2;
-  int value1 = (col1->red << 16) | (col1->green << 8) | col1->blue;
-  int value2 = (col2->red << 16) | (col2->green << 8) | col2->blue;
+  int value1 = (col1->gfc_red << 16) | (col1->gfc_green << 8) | col1->gfc_blue;
+  int value2 = (col2->gfc_red << 16) | (col2->gfc_green << 8) | col2->gfc_blue;
   return value1 - value2;
 }
 
@@ -1123,7 +1152,8 @@ create_new_image_data(Gif_Stream *gfs, int optimize_flags)
 
     /* save previous data if necessary */
     if (cur_gfi->disposal == GIF_DISPOSAL_PREVIOUS) {
-      previous_data = Gif_NewArray(uint16_t, screen_size);
+      if (!previous_data)
+          previous_data = Gif_NewArray(uint16_t, screen_size);
       copy_data_area(previous_data, this_data, cur_gfi);
     }
 
@@ -1178,16 +1208,17 @@ create_new_image_data(Gif_Stream *gfs, int optimize_flags)
       copy_data_area(last_data, this_data, cur_gfi);
     else if (cur_gfi->disposal == GIF_DISPOSAL_BACKGROUND)
       fill_data_area(last_data, background, cur_gfi);
-    else
+    else if (cur_gfi->disposal != GIF_DISPOSAL_PREVIOUS)
       assert(0 && "optimized frame has strange disposal");
 
     if (cur_unopt_gfi.disposal == GIF_DISPOSAL_BACKGROUND)
       fill_data_area(this_data, background, &cur_unopt_gfi);
-    else if (cur_unopt_gfi.disposal == GIF_DISPOSAL_PREVIOUS) {
+    else if (cur_unopt_gfi.disposal == GIF_DISPOSAL_PREVIOUS)
       copy_data_area(this_data, previous_data, &cur_unopt_gfi);
-      Gif_DeleteArray(previous_data);
-    }
   }
+
+  if (previous_data)
+      Gif_DeleteArray(previous_data);
 }
 
 
@@ -1205,9 +1236,9 @@ initialize_optimizer(Gif_Stream *gfs)
 
   /* combine colormaps */
   all_colormap = Gif_NewFullColormap(1, 384);
-  all_colormap->col[0].red = 255;
-  all_colormap->col[0].green = 255;
-  all_colormap->col[0].blue = 255;
+  all_colormap->col[0].gfc_red = 255;
+  all_colormap->col[0].gfc_green = 255;
+  all_colormap->col[0].gfc_blue = 255;
 
   in_global_map = gfs->global;
   if (!in_global_map) {
@@ -1215,7 +1246,7 @@ initialize_optimizer(Gif_Stream *gfs)
     in_global_map = Gif_NewFullColormap(256, 256);
     col = in_global_map->col;
     for (i = 0; i < 256; i++, col++)
-      col->red = col->green = col->blue = i;
+      col->gfc_red = col->gfc_green = col->gfc_blue = i;
   }
 
   {
