@@ -12,6 +12,29 @@
 #include <assert.h>
 #include <string.h>
 
+static inline uint32_t rgb_distance(const Gif_Color* x, const Gif_Color* y) {
+    return (x->gfc_red - y->gfc_red) * (x->gfc_red - y->gfc_red)
+        + (x->gfc_green - y->gfc_green) * (x->gfc_green - y->gfc_green)
+        + (x->gfc_blue - y->gfc_blue) * (x->gfc_blue - y->gfc_blue);
+}
+
+static inline unsigned kd3_distance(const int x[3], const int y[3]) {
+    return (x[0] - y[0]) * (x[0] - y[0])
+        + (x[1] - y[1]) * (x[1] - y[1])
+        + (x[2] - y[2]) * (x[2] - y[2]);
+}
+
+static inline void kd3_luminance_transform(int x[3]) {
+    /* For grayscale colormaps, use distance in luminance space instead of
+       distance in RGB space. The weights for the R,G,B components in
+       luminance space are 0.299,0.587,0.114. Using the proportional factors
+       306, 601, and 117 we get a scaled gray value between 0 and 255 *
+       1024. Thanks to Christian Kumpf, <kumpf@igd.fhg.de>, for providing a
+       patch. */
+    int gray = (306 * x[0] + 601 * x[1] + 117 * x[2]) >> 10;
+    x[0] = x[1] = x[2] = gray;
+}
+
 typedef struct Gif_Histogram {
   Gif_Color *c;
   int n;
@@ -48,14 +71,13 @@ add_histogram_color(Gif_Color *color, Gif_Histogram *hist, unsigned long count)
 {
   Gif_Color *hc = hist->c;
   int hcap = hist->cap - 1;
-  int i = (((color->red & 0xF0) << 4) | (color->green & 0xF0)
-	   | (color->blue >> 4)) & hcap;
-  int hash2 = ((((color->red & 0x0F) << 8) | ((color->green & 0x0F) << 4)
-		| (color->blue & 0x0F)) & hcap) | 1;
+  int i = (((color->gfc_red & 0xF0) << 4) | (color->gfc_green & 0xF0)
+	   | (color->gfc_blue >> 4)) & hcap;
+  int hash2 = ((((color->gfc_red & 0x0F) << 8) | ((color->gfc_green & 0x0F) << 4)
+		| (color->gfc_blue & 0x0F)) & hcap) | 1;
 
   for (; hc[i].haspixel; i = (i + hash2) & hcap)
-    if (hc[i].red == color->red && hc[i].green == color->green
-	&& hc[i].blue == color->blue) {
+    if (GIF_COLOREQ(&hc[i], color)) {
       hc[i].pixel += count;
       color->haspixel = 1;
       color->pixel = i;
@@ -204,7 +226,7 @@ red_sort_compare(const void *va, const void *vb)
 {
   const Gif_Color *a = (const Gif_Color *)va;
   const Gif_Color *b = (const Gif_Color *)vb;
-  return a->red - b->red;
+  return a->gfc_red - b->gfc_red;
 }
 
 static int
@@ -212,7 +234,7 @@ green_sort_compare(const void *va, const void *vb)
 {
   const Gif_Color *a = (const Gif_Color *)va;
   const Gif_Color *b = (const Gif_Color *)vb;
-  return a->green - b->green;
+  return a->gfc_green - b->gfc_green;
 }
 
 static int
@@ -220,7 +242,7 @@ blue_sort_compare(const void *va, const void *vb)
 {
   const Gif_Color *a = (const Gif_Color *)va;
   const Gif_Color *b = (const Gif_Color *)vb;
-  return a->blue - b->blue;
+  return a->gfc_blue - b->gfc_blue;
 }
 
 
@@ -243,7 +265,7 @@ typedef struct {
 } adaptive_slot;
 
 Gif_Colormap *
-colormap_median_cut(Gif_Color *hist, int nhist, int adapt_size)
+colormap_median_cut(Gif_Color *hist, int nhist, int adapt_size, int complain)
 {
   adaptive_slot *slots = Gif_NewArray(adaptive_slot, adapt_size);
   Gif_Colormap *gfcm = Gif_NewFullColormap(adapt_size, 256);
@@ -256,10 +278,10 @@ colormap_median_cut(Gif_Color *hist, int nhist, int adapt_size)
 
   if (adapt_size < 2 || adapt_size > 256)
     fatal_error("adaptive palette size must be between 2 and 256");
-  if (adapt_size >= nhist) {
+  if (adapt_size >= nhist && complain)
     warning(1, "trivial adaptive palette (only %d colors in source)", nhist);
+  if (adapt_size >= nhist)
     adapt_size = nhist;
-  }
 
   /* 0. remove any transparent color from consideration; reduce adaptive
      palette size to accommodate transparency if it looks like that'll be
@@ -307,21 +329,21 @@ colormap_median_cut(Gif_Color *hist, int nhist, int adapt_size)
       Gif_Color *trav = slice;
       minc = maxc = *trav;
       for (i = 1, trav++; i < split->count; i++, trav++) {
-	minc.red = min(minc.red, trav->red);
-	maxc.red = max(maxc.red, trav->red);
-	minc.green = min(minc.green, trav->green);
-	maxc.green = max(maxc.green, trav->green);
-	minc.blue = min(minc.blue, trav->blue);
-	maxc.blue = max(maxc.blue, trav->blue);
+	minc.gfc_red = min(minc.gfc_red, trav->gfc_red);
+	maxc.gfc_red = max(maxc.gfc_red, trav->gfc_red);
+	minc.gfc_green = min(minc.gfc_green, trav->gfc_green);
+	maxc.gfc_green = max(maxc.gfc_green, trav->gfc_green);
+	minc.gfc_blue = min(minc.gfc_blue, trav->gfc_blue);
+	maxc.gfc_blue = max(maxc.gfc_blue, trav->gfc_blue);
       }
     }
 
     /* 2.3. decide how to split it. use the luminance method. also sort the
        colors. */
     {
-      double red_diff = 0.299 * (maxc.red - minc.red);
-      double green_diff = 0.587 * (maxc.green - minc.green);
-      double blue_diff = 0.114 * (maxc.blue - minc.blue);
+      double red_diff = 0.299 * (maxc.gfc_red - minc.gfc_red);
+      double green_diff = 0.587 * (maxc.gfc_green - minc.gfc_green);
+      double blue_diff = 0.114 * (maxc.gfc_blue - minc.gfc_blue);
       if (red_diff >= green_diff && red_diff >= blue_diff)
 	qsort(slice, split->count, sizeof(Gif_Color), red_sort_compare);
       else if (green_diff >= blue_diff)
@@ -361,13 +383,13 @@ colormap_median_cut(Gif_Color *hist, int nhist, int adapt_size)
     double red_total = 0, green_total = 0, blue_total = 0;
     Gif_Color *slice = &hist[ slots[i].first ];
     for (j = 0; j < slots[i].count; j++) {
-      red_total += slice[j].red * slice[j].pixel;
-      green_total += slice[j].green * slice[j].pixel;
-      blue_total += slice[j].blue * slice[j].pixel;
+      red_total += slice[j].gfc_red * slice[j].pixel;
+      green_total += slice[j].gfc_green * slice[j].pixel;
+      blue_total += slice[j].gfc_blue * slice[j].pixel;
     }
-    adapt[i].red = (uint8_t)(red_total / slots[i].pixel);
-    adapt[i].green = (uint8_t)(green_total / slots[i].pixel);
-    adapt[i].blue = (uint8_t)(blue_total / slots[i].pixel);
+    adapt[i].gfc_red = (uint8_t)(red_total / slots[i].pixel);
+    adapt[i].gfc_green = (uint8_t)(green_total / slots[i].pixel);
+    adapt[i].gfc_blue = (uint8_t)(blue_total / slots[i].pixel);
     adapt[i].haspixel = 0;
   }
 
@@ -379,7 +401,8 @@ colormap_median_cut(Gif_Color *hist, int nhist, int adapt_size)
 
 
 static Gif_Colormap *
-colormap_diversity(Gif_Color *hist, int nhist, int adapt_size, int blend)
+colormap_diversity(Gif_Color *hist, int nhist, int adapt_size, int complain,
+                   int blend)
 {
   uint32_t *min_dist = Gif_NewArray(uint32_t, nhist);
   int *closest = Gif_NewArray(int, nhist);
@@ -394,10 +417,10 @@ colormap_diversity(Gif_Color *hist, int nhist, int adapt_size, int blend)
 
   if (adapt_size < 2 || adapt_size > 256)
     fatal_error("adaptive palette size must be between 2 and 256");
-  if (adapt_size > nhist) {
+  if (adapt_size > nhist && complain)
     warning(1, "trivial adaptive palette (only %d colors in source)", nhist);
+  if (adapt_size > nhist)
     adapt_size = nhist;
-  }
 
   /* 0. remove any transparent color from consideration; reduce adaptive
      palette size to accommodate transparency if it looks like that'll be
@@ -428,24 +451,22 @@ colormap_diversity(Gif_Color *hist, int nhist, int adapt_size, int blend)
 
   /* 2. choose colors one at a time */
   for (nadapt = 0; nadapt < adapt_size; nadapt++) {
-    int chosen = 0;
+    int chosen;
+
+    /* 2.0. find the first non-chosen color */
+    for (chosen = 0; min_dist[chosen] == 0; ++chosen)
+        /* do nothing */;
 
     /* 2.1. choose the color to be added */
     if (nadapt == 0 || (nadapt >= 10 && nadapt % 2 == 0)) {
-      /* 2.1a. choose based on popularity from unchosen colors; we've sorted
-	 them on popularity, so just choose the first in the list */
-      for (; chosen < nhist; chosen++)
-	if (min_dist[chosen])
-	  break;
+        /* 2.1a. want most popular unchosen color; we've sorted them on
+           popularity, so we're done! */
 
     } else {
-      /* 2.1b. choose based on diversity from unchosen colors */
-      uint32_t chosen_dist = 0;
-      for (i = 0; i < nhist; i++)
-	if (min_dist[i] > chosen_dist) {
-	  chosen = i;
-	  chosen_dist = min_dist[i];
-	}
+        /* 2.1b. choose based on diversity from unchosen colors */
+        for (i = chosen + 1; i != nhist; ++i)
+            if (min_dist[i] > min_dist[chosen])
+                chosen = i;
     }
 
     /* 2.2. add the color */
@@ -453,21 +474,14 @@ colormap_diversity(Gif_Color *hist, int nhist, int adapt_size, int blend)
     closest[chosen] = nadapt;
 
     /* 2.3. adjust the min_dist array */
-    {
-      int red = hist[chosen].red, green = hist[chosen].green,
-	blue = hist[chosen].blue;
-      Gif_Color *h = hist;
-      for (i = 0; i < nhist; i++, h++)
-	if (min_dist[i]) {
-	  uint32_t dist = (h->red - red) * (h->red - red)
-	    + (h->green - green) * (h->green - green)
-	    + (h->blue - blue) * (h->blue - blue);
-	  if (dist < min_dist[i]) {
-	    min_dist[i] = dist;
-	    closest[i] = nadapt;
-	  }
-	}
-    }
+    for (i = 0; i < nhist; ++i)
+        if (min_dist[i]) {
+            uint32_t dist = rgb_distance(&hist[i], &hist[chosen]);
+            if (dist < min_dist[i]) {
+                min_dist[i] = dist;
+                closest[i] = nadapt;
+            }
+        }
   }
 
   /* 3. make the new palette by choosing one color from each slot. */
@@ -487,9 +501,9 @@ colormap_diversity(Gif_Color *hist, int nhist, int adapt_size, int blend)
       for (j = 0; j < nhist; j++)
 	if (closest[j] == i) {
 	  uint32_t pixel = hist[j].pixel;
-	  red_total += hist[j].red * pixel;
-	  green_total += hist[j].green * pixel;
-	  blue_total += hist[j].blue * pixel;
+	  red_total += hist[j].gfc_red * pixel;
+	  green_total += hist[j].gfc_green * pixel;
+	  blue_total += hist[j].gfc_blue * pixel;
 	  pixel_total += pixel;
 	  if (min_dist[j])
 	    mismatch_pixel_total += pixel;
@@ -504,13 +518,13 @@ colormap_diversity(Gif_Color *hist, int nhist, int adapt_size, int blend)
 	/* Favor, by a smallish amount, the color the plain diversity
            algorithm would pick. */
 	uint32_t pixel = hist[match].pixel * 2;
-	red_total += hist[match].red * pixel;
-	green_total += hist[match].green * pixel;
-	blue_total += hist[match].blue * pixel;
+	red_total += hist[match].gfc_red * pixel;
+	green_total += hist[match].gfc_green * pixel;
+	blue_total += hist[match].gfc_blue * pixel;
 	pixel_total += pixel;
-	adapt[i].red = (uint8_t)(red_total / pixel_total);
-	adapt[i].green = (uint8_t)(green_total / pixel_total);
-	adapt[i].blue = (uint8_t)(blue_total / pixel_total);
+	adapt[i].gfc_red = (uint8_t)(red_total / pixel_total);
+	adapt[i].gfc_green = (uint8_t)(green_total / pixel_total);
+	adapt[i].gfc_blue = (uint8_t)(blue_total / pixel_total);
       }
       adapt[i].haspixel = 0;
     }
@@ -524,167 +538,302 @@ colormap_diversity(Gif_Color *hist, int nhist, int adapt_size, int blend)
 
 
 Gif_Colormap *
-colormap_blend_diversity(Gif_Color *hist, int nhist, int adapt_size)
+colormap_blend_diversity(Gif_Color *hist, int nhist, int adapt_size,
+                         int complain)
 {
-  return colormap_diversity(hist, nhist, adapt_size, 1);
+    return colormap_diversity(hist, nhist, adapt_size, complain, 1);
 }
 
 Gif_Colormap *
-colormap_flat_diversity(Gif_Color *hist, int nhist, int adapt_size)
+colormap_flat_diversity(Gif_Color *hist, int nhist, int adapt_size,
+                        int complain)
 {
-  return colormap_diversity(hist, nhist, adapt_size, 0);
+    return colormap_diversity(hist, nhist, adapt_size, complain, 0);
 }
 
 
-struct color_hash_item {
-  uint8_t red;
-  uint8_t green;
-  uint8_t blue;
-  uint32_t pixel;
-  color_hash_item *next;
-};
-#define COLOR_HASH_SIZE 20023
-#define COLOR_HASH_CODE(r, g, b)	((uint32_t)(r * 33023 + g * 30013 + b * 27011) % COLOR_HASH_SIZE)
-
-
 /*****
- * color_hash_item allocation and deallocation
+ * kd_tree allocation and deallocation
  **/
 
-static color_hash_item *hash_item_alloc_list;
-static int hash_item_alloc_left;
-#define HASH_ITEM_ALLOC_AMOUNT 512
+typedef struct kd3_item {
+    int a[3];
+    int index;
+} kd3_item;
 
-static color_hash_item **
-new_color_hash(void)
-{
-  int i;
-  color_hash_item **hash = Gif_NewArray(color_hash_item *, COLOR_HASH_SIZE);
-  for (i = 0; i < COLOR_HASH_SIZE; i++)
-    hash[i] = 0;
-  return hash;
+typedef struct kd3_treepos {
+    int pivot;
+    int offset;
+} kd3_treepos;
+
+typedef struct kd3_tree {
+    kd3_treepos* tree;
+    int ntree;
+    int disabled;
+    kd3_item* items;
+    int nitems;
+    int items_cap;
+    int maxdepth;
+    unsigned (*distance)(const int[3], const int[3]);
+    void (*transform)(int[3]);
+    unsigned* xradius;
+} kd3_tree;
+
+void kd3_init(kd3_tree* kd3, unsigned (*distance)(const int[3], const int[3]),
+              void (*transform)(int[3])) {
+    kd3->tree = NULL;
+    kd3->items = Gif_NewArray(kd3_item, 256);
+    kd3->nitems = 0;
+    kd3->items_cap = 256;
+    kd3->distance = distance;
+    kd3->transform = transform;
+    kd3->xradius = NULL;
+    kd3->disabled = -1;
 }
 
-
-static color_hash_item *
-new_color_hash_item(uint8_t red, uint8_t green, uint8_t blue)
-{
-  color_hash_item *chi;
-  if (hash_item_alloc_left <= 0) {
-    color_hash_item *new_alloc =
-      Gif_NewArray(color_hash_item, HASH_ITEM_ALLOC_AMOUNT);
-    new_alloc[HASH_ITEM_ALLOC_AMOUNT-1].next = hash_item_alloc_list;
-    hash_item_alloc_list = new_alloc;
-    hash_item_alloc_left = HASH_ITEM_ALLOC_AMOUNT - 1;
-  }
-
-  --hash_item_alloc_left;
-  chi = &hash_item_alloc_list[hash_item_alloc_left];
-  chi->red = red;
-  chi->green = green;
-  chi->blue = blue;
-  chi->next = 0;
-  return chi;
+void kd3_cleanup(kd3_tree* kd3) {
+    Gif_DeleteArray(kd3->tree);
+    Gif_DeleteArray(kd3->items);
+    Gif_DeleteArray(kd3->xradius);
 }
 
-static void
-free_all_color_hash_items(void)
-{
-  while (hash_item_alloc_list) {
-    color_hash_item *next =
-      hash_item_alloc_list[HASH_ITEM_ALLOC_AMOUNT - 1].next;
-    Gif_DeleteArray(hash_item_alloc_list);
-    hash_item_alloc_list = next;
-  }
-  hash_item_alloc_left = 0;
+void kd3_add(kd3_tree* kd3, int a0, int a1, int a2) {
+    assert(!kd3->tree);
+    if (kd3->nitems == kd3->items_cap) {
+        kd3->items_cap *= 2;
+        Gif_ReArray(kd3->items, kd3_item, kd3->items_cap);
+    }
+    kd3->items[kd3->nitems].a[0] = a0;
+    kd3->items[kd3->nitems].a[1] = a1;
+    kd3->items[kd3->nitems].a[2] = a2;
+    if (kd3->transform)
+        kd3->transform(kd3->items[kd3->nitems].a);
+    kd3->items[kd3->nitems].index = kd3->nitems;
+    ++kd3->nitems;
 }
 
+void kd3_add_8to15(kd3_tree* kd3, int a0, int a1, int a2) {
+    kd3_add(kd3, (a0 << 7) | (a0 >> 1), (a1 << 7) | (a1 >> 1),
+            (a2 << 7) | (a2 >> 1));
+}
 
-static int
-hash_color(int red, int green, int blue,
-	   color_hash_item **hash, Gif_Colormap *new_cm)
-{
-  uint32_t hash_code = COLOR_HASH_CODE(red, green, blue);
-  color_hash_item *prev = 0, *trav;
+static int kd3_item_compar_0(const void* a, const void* b) {
+    const kd3_item* aa = (const kd3_item*) a, *bb = (const kd3_item*) b;
+    return aa->a[0] - bb->a[0];
+}
 
-  /* Is new_cm grayscale? We cache the answer here. */
-  static Gif_Colormap *cached_new_cm;
-  static int new_cm_grayscale;
+static int kd3_item_compar_1(const void* a, const void* b) {
+    const kd3_item* aa = (const kd3_item*) a, *bb = (const kd3_item*) b;
+    return aa->a[1] - bb->a[1];
+}
 
-  for (trav = hash[hash_code]; trav; prev = trav, trav = trav->next)
-    if (trav->red == red && trav->green == green && trav->blue == blue)
-      return trav->pixel;
+static int kd3_item_compar_2(const void* a, const void* b) {
+    const kd3_item* aa = (const kd3_item*) a, *bb = (const kd3_item*) b;
+    return aa->a[2] - bb->a[2];
+}
 
-  trav = new_color_hash_item(red, green, blue);
-  if (prev)
-    prev->next = trav;
-  else
-    hash[hash_code] = trav;
+static int (*kd3_item_compars[])(const void*, const void*) = {
+    &kd3_item_compar_0, &kd3_item_compar_1, &kd3_item_compar_2
+};
 
-  /* calculate whether new_cm is grayscale */
-  if (new_cm != cached_new_cm) {
-    int i;
-    Gif_Color *col = new_cm->col;
-    cached_new_cm = new_cm;
-    new_cm_grayscale = 1;
-    for (i = 0; i < new_cm->ncol && new_cm_grayscale; i++)
-      if (col[i].red != col[i].green || col[i].green != col[i].blue
-	  || col[i].blue != col[i].red)
-	new_cm_grayscale = 0;
-  }
-
-  /* find the closest color in the new colormap */
-  {
-    Gif_Color *col = new_cm->col;
-    int ncol = new_cm->ncol, i, found;
-    uint32_t min_dist = 0xFFFFFFFFU;
-
-    if (new_cm_grayscale) {
-      /* If the new colormap is 100% grayscale, then use distance in luminance
-	 space instead of distance in RGB space. The weights for the R,G,B
-	 components in luminance space are 0.299,0.587,0.114. We calculate a
-	 gray value for the input color first and compare that against the
-	 available grays in the colormap. Thanks to Christian Kumpf,
-	 <kumpf@igd.fhg.de>, for providing a patch.
-
-	 Note on the calculation of 'gray': Using the factors 306, 601, and
-	 117 (proportional to 0.299,0.587,0.114) we get a scaled gray value
-	 between 0 and 255 * 1024. */
-      int gray = 306 * red + 601 * green + 117 * blue;
-      for (i = 0; i < ncol; i++)
-	if (col[i].haspixel != 255) {
-	  int in_gray = 1024 * col[i].red;
-	  uint32_t dist = abs(gray - in_gray);
-	  if (dist < min_dist) {
-	    min_dist = dist;
-	    found = i;
-	  }
-	}
-
-    } else {
-      /* Use straight-line Euclidean distance in RGB space */
-      for (i = 0; i < ncol; i++)
-	if (col[i].haspixel != 255) {
-	  uint32_t dist = (red - col[i].red) * (red - col[i].red)
-	    + (green - col[i].green) * (green - col[i].green)
-	    + (blue - col[i].blue) * (blue - col[i].blue);
-	  if (dist < min_dist) {
-	    min_dist = dist;
-	    found = i;
-	  }
-	}
+static int kd3_build_range(kd3_tree* kd3, kd3_item* items,
+                           int l, int r, int n, int depth) {
+    int m, aindex = depth % 3;
+    if (depth > kd3->maxdepth)
+        kd3->maxdepth = depth;
+    while (n >= kd3->ntree) {
+        kd3->ntree *= 2;
+        Gif_ReArray(kd3->tree, kd3_treepos, kd3->ntree);
+    }
+    if (l + 1 >= r) {
+        kd3->tree[n].pivot = (l == r ? -1 : items[l].index);
+        kd3->tree[n].offset = -1;
+        return 2;
     }
 
-    trav->pixel = found;
-    return found;
-  }
+    qsort(&items[l], r - l, sizeof(kd3_item), kd3_item_compars[aindex]);
+    m = l + ((r - l) >> 1);
+    while (m > l && items[m].a[aindex] == items[m-1].a[aindex])
+        --m;
+    if (m == l)
+        kd3->tree[n].pivot = items[m].a[aindex];
+    else
+        kd3->tree[n].pivot = items[m-1].a[aindex]
+            + ((items[m].a[aindex] - items[m-1].a[aindex]) >> 1);
+    int nl = kd3_build_range(kd3, items, l, m, n+1, depth+1);
+    kd3->tree[n].offset = 1+nl;
+    int nr = kd3_build_range(kd3, items, m, r, n+1+nl, depth+1);
+    return 1+nl+nr;
+}
+
+static int kd3_item_all_compar(const void* a, const void* b) {
+    const kd3_item* aa = (const kd3_item*) a, *bb = (const kd3_item*) b;
+    if (aa->a[0] - bb->a[0])
+        return aa->a[0] - bb->a[0];
+    else if (aa->a[1] - bb->a[1])
+        return aa->a[1] - bb->a[1];
+    else
+        return aa->a[2] - bb->a[2];
+}
+
+void kd3_print(kd3_tree* kd3, int depth, kd3_treepos* p, int* a, int* b) {
+    int i;
+    char x[10][6];
+    for (i = 0; i != 3; ++i) {
+        if (a[i] == 0x80000000)
+            sprintf(x[2*i], "*");
+        else
+            sprintf(x[2*i], "%d", a[i]);
+        if (b[i] == 0x7FFFFFFF)
+            sprintf(x[2*i+1], "*");
+        else
+            sprintf(x[2*i+1], "%d", b[i]);
+    }
+    printf("%*s<%s:%s,%s:%s,%s:%s>", depth*3, "",
+           x[0], x[1], x[2], x[3], x[4], x[5]);
+    if (p->offset < 0) {
+        if (p->pivot >= 0) {
+            assert(kd3->items[p->pivot].a[0] >= a[0]);
+            assert(kd3->items[p->pivot].a[1] >= a[1]);
+            assert(kd3->items[p->pivot].a[2] >= a[2]);
+            assert(kd3->items[p->pivot].a[0] < b[0]);
+            assert(kd3->items[p->pivot].a[1] < b[1]);
+            assert(kd3->items[p->pivot].a[2] < b[2]);
+            printf(" ** @%d: <%d,%d,%d>\n", p->pivot, kd3->items[p->pivot].a[0], kd3->items[p->pivot].a[1], kd3->items[p->pivot].a[2]);
+        }
+    } else {
+        int aindex = depth % 3;
+        assert(p->pivot >= a[aindex]);
+        assert(p->pivot < b[aindex]);
+        printf((aindex == 0 ? " | <%d,_,_>\n" :
+                aindex == 1 ? " | <_,%d,_>\n" : " | <_,_,%d>\n"), p->pivot);
+        int x[3];
+        memcpy(x, b, sizeof(int) * 3);
+        x[aindex] = p->pivot;
+        kd3_print(kd3, depth + 1, p + 1, a, x);
+        memcpy(x, a, sizeof(int) * 3);
+        x[aindex] = p->pivot;
+        kd3_print(kd3, depth + 1, p + p->offset, x, b);
+    }
+}
+
+void kd3_build(kd3_tree* kd3) {
+    kd3_item* items;
+    int i, j, delta;
+    assert(!kd3->tree);
+
+    // create xradius
+    kd3->xradius = Gif_NewArray(unsigned, kd3->nitems);
+    for (i = 0; i != kd3->nitems; ++i)
+        kd3->xradius[i] = (unsigned) -1;
+    for (i = 0; i != kd3->nitems; ++i)
+        for (j = i + 1; j != kd3->nitems; ++j) {
+            unsigned dist = kd3->distance(kd3->items[i].a, kd3->items[j].a);
+            unsigned radius = dist / 4;
+            if (radius < kd3->xradius[i])
+                kd3->xradius[i] = radius;
+            if (radius < kd3->xradius[j])
+                kd3->xradius[j] = radius;
+        }
+
+    // create tree
+    kd3->tree = Gif_NewArray(kd3_treepos, 256);
+    kd3->ntree = 256;
+    kd3->maxdepth = 0;
+
+    // create copy of items; remove duplicates
+    items = Gif_NewArray(kd3_item, kd3->nitems);
+    memcpy(items, kd3->items, sizeof(kd3_item) * kd3->nitems);
+    qsort(items, kd3->nitems, sizeof(kd3_item), kd3_item_all_compar);
+    for (i = 0, delta = 1; i != kd3->nitems - delta; ++i)
+        if (items[i].a[0] == items[i+delta].a[0]
+            && items[i].a[1] == items[i+delta].a[1]
+            && items[i].a[2] == items[i+delta].a[2])
+            ++delta, --i;
+        else if (delta > 1)
+            items[i+1] = items[i+delta];
+
+    kd3_build_range(kd3, items, 0, kd3->nitems - (delta - 1), 0, 0);
+    assert(kd3->maxdepth < 32);
+
+    Gif_DeleteArray(items);
+}
+
+void kd3_disable(kd3_tree* kd3, int i) {
+    assert(kd3->disabled < 0);
+    kd3->disabled = i;
+}
+
+void kd3_enable_all(kd3_tree* kd3) {
+    kd3->disabled = -1;
+}
+
+int kd3_closest_transformed(const kd3_tree* kd3, const int a[3]) {
+    assert(kd3->tree);
+    const kd3_treepos* stack[32];
+    uint8_t state[32];
+    int stackpos = 0;
+    int result = -1;
+    unsigned mindist = (unsigned) -1;
+    stack[0] = kd3->tree;
+    state[0] = 0;
+
+    while (stackpos >= 0) {
+        assert(stackpos < 32);
+        const kd3_treepos* p = stack[stackpos];
+
+        if (p->offset < 0) {
+            if (p->pivot >= 0 && kd3->disabled != p->pivot) {
+                unsigned dist = kd3->distance(kd3->items[p->pivot].a, a);
+                if (dist < mindist) {
+                    mindist = dist;
+                    result = p->pivot;
+                }
+            }
+            if (--stackpos >= 0)
+                ++state[stackpos];
+        } else if (state[stackpos] == 0) {
+            if (a[stackpos % 3] < p->pivot)
+                stack[stackpos + 1] = p + 1;
+            else
+                stack[stackpos + 1] = p + p->offset;
+            ++stackpos;
+            state[stackpos] = 0;
+        } else {
+            int delta = a[stackpos % 3] - p->pivot;
+            if (state[stackpos] == 1
+                && (unsigned) (delta * delta) < mindist) {
+                if (delta < 0)
+                    stack[stackpos + 1] = p + p->offset;
+                else
+                    stack[stackpos + 1] = p + 1;
+                ++stackpos;
+                state[stackpos] = 0;
+            } else if (--stackpos >= 0)
+                ++state[stackpos];
+        }
+    }
+
+    return result;
+}
+
+int kd3_closest(const kd3_tree* kd3, int a0, int a1, int a2) {
+    int a[3] = {a0, a1, a2};
+    if (kd3->transform)
+        kd3->transform(a);
+    return kd3_closest_transformed(kd3, a);
+}
+
+int kd3_closest_8to15(const kd3_tree* kd3, int a0, int a1, int a2) {
+    return kd3_closest(kd3, (a0 << 7) | (a0 >> 1), (a1 << 7) | (a1 >> 1),
+                       (a2 << 7) | (a2 >> 1));
 }
 
 
 void
 colormap_image_posterize(Gif_Image *gfi, uint8_t *new_data,
-			 Gif_Colormap *old_cm, Gif_Colormap *new_cm,
-			 color_hash_item **hash, uint32_t *histogram)
+			 Gif_Colormap *old_cm, kd3_tree* kd3,
+			 uint32_t *histogram)
 {
   int ncol = old_cm->ncol;
   Gif_Color *col = old_cm->col;
@@ -693,14 +842,10 @@ colormap_image_posterize(Gif_Image *gfi, uint8_t *new_data,
   int transparent = gfi->transparent;
 
   /* find closest colors in new colormap */
-  for (i = 0; i < ncol; i++)
-    if (col[i].haspixel)
-      map[i] = col[i].pixel;
-    else {
-      map[i] = col[i].pixel =
-	hash_color(col[i].red, col[i].green, col[i].blue, hash, new_cm);
+  for (i = 0; i < ncol; i++) {
+      map[i] = col[i].pixel = kd3_closest_8to15(kd3, col[i].gfc_red, col[i].gfc_green, col[i].gfc_blue);
       col[i].haspixel = 1;
-    }
+  }
 
   /* map image */
   for (j = 0; j < gfi->height; j++) {
@@ -715,35 +860,42 @@ colormap_image_posterize(Gif_Image *gfi, uint8_t *new_data,
 
 
 #define DITHER_SCALE	1024
+#define DITHER_SHIFT    10
 #define DITHER_SCALE_M1	(DITHER_SCALE-1)
+#define DITHER_ITEM2ERR (1<<(DITHER_SHIFT-7))
 #define N_RANDOM_VALUES	512
 
+typedef struct color_erritem {
+    int32_t a[3];
+} color_erritem;
+
 void
 colormap_image_floyd_steinberg(Gif_Image *gfi, uint8_t *all_new_data,
-			       Gif_Colormap *old_cm, Gif_Colormap *new_cm,
-			       color_hash_item **hash, uint32_t *histogram)
+			       Gif_Colormap *old_cm, kd3_tree* kd3,
+                               uint32_t *histogram)
 {
   static int32_t *random_values = 0;
 
   int width = gfi->width;
   int dither_direction = 0;
   int transparent = gfi->transparent;
-  int i, j;
-  int32_t *r_err, *g_err, *b_err, *r_err1, *g_err1, *b_err1;
-  Gif_Color *col = old_cm->col;
-  Gif_Color *new_col = new_cm->col;
+  int i, j, k;
+  color_erritem *err, *err1;
+
+  /* Initialize distances */
+  for (i = 0; i < old_cm->ncol; ++i) {
+      Gif_Color* c = &old_cm->col[i];
+      c->pixel = kd3_closest_8to15(kd3, c->gfc_red, c->gfc_green, c->gfc_blue);
+      c->haspixel = 1;
+  }
 
   /* This code was written with reference to ppmquant by Jef Poskanzer, part
      of the pbmplus package. */
 
   /* Initialize Floyd-Steinberg error vectors to small random values, so we
      don't get artifacts on the top row */
-  r_err = Gif_NewArray(int32_t, width + 2);
-  g_err = Gif_NewArray(int32_t, width + 2);
-  b_err = Gif_NewArray(int32_t, width + 2);
-  r_err1 = Gif_NewArray(int32_t, width + 2);
-  g_err1 = Gif_NewArray(int32_t, width + 2);
-  b_err1 = Gif_NewArray(int32_t, width + 2);
+  err = Gif_NewArray(color_erritem, width + 2);
+  err1 = Gif_NewArray(color_erritem, width + 2);
   /* Use the same random values on each call in an attempt to minimize
      "jumping dithering" effects on animations */
   if (!random_values) {
@@ -752,12 +904,11 @@ colormap_image_floyd_steinberg(Gif_Image *gfi, uint8_t *all_new_data,
       random_values[i] = RANDOM() % (DITHER_SCALE_M1 * 2) - DITHER_SCALE_M1;
   }
   for (i = 0; i < gfi->width + 2; i++) {
-    int j = (i + gfi->left) * 3;
-    r_err[i] = random_values[ (j + 0) % N_RANDOM_VALUES ];
-    g_err[i] = random_values[ (j + 1) % N_RANDOM_VALUES ];
-    b_err[i] = random_values[ (j + 2) % N_RANDOM_VALUES ];
+    j = (i + gfi->left) * 3;
+    for (k = 0; k < 3; ++k)
+        err[i].a[k] = random_values[ (j + k) % N_RANDOM_VALUES ];
   }
-  /* *_err1 initialized below */
+  /* err1 initialized below */
 
   /* Do the image! */
   for (j = 0; j < gfi->height; j++) {
@@ -776,53 +927,48 @@ colormap_image_floyd_steinberg(Gif_Image *gfi, uint8_t *all_new_data,
     new_data = all_new_data + j * width + x;
 
     for (i = 0; i < width + 2; i++)
-      r_err1[i] = g_err1[i] = b_err1[i] = 0;
+      err1[i].a[0] = err1[i].a[1] = err1[i].a[2] = 0;
 
     /* Do a single row */
     while (x >= 0 && x < width) {
-      int e, use_r, use_g, use_b;
+      int e;
+      color_erritem use, usexf;
 
       /* the transparent color never gets adjusted */
       if (*data == transparent)
 	goto next;
 
       /* use Floyd-Steinberg errors to adjust actual color */
-      use_r = col[*data].red + r_err[x+1] / DITHER_SCALE;
-      use_g = col[*data].green + g_err[x+1] / DITHER_SCALE;
-      use_b = col[*data].blue + b_err[x+1] / DITHER_SCALE;
-      use_r = max(use_r, 0);  use_r = min(use_r, 255);
-      use_g = max(use_g, 0);  use_g = min(use_g, 255);
-      use_b = max(use_b, 0);  use_b = min(use_b, 255);
+      for (k = 0; k < 3; ++k) {
+          use.a[k] = old_cm->col[*data].gfc_array[k];
+          use.a[k] = (use.a[k] << 7) | (use.a[k] >> 1);
+          use.a[k] += (err[x+1].a[k] & ~(DITHER_ITEM2ERR-1)) / DITHER_ITEM2ERR;
+          use.a[k] = max(use.a[k], 0);
+          use.a[k] = min(use.a[k], (255 << 7) | (255 >> 1));
+      }
 
-      *new_data = hash_color(use_r, use_g, use_b, hash, new_cm);
+      usexf = use;
+      if (kd3->transform)
+          kd3->transform(usexf.a);
+      e = old_cm->col[*data].pixel;
+      if (kd3->distance(kd3->items[e].a, usexf.a) < kd3->xradius[e])
+          *new_data = e;
+      else
+          *new_data = kd3_closest_transformed(kd3, usexf.a);
       histogram[*new_data]++;
 
       /* calculate and propagate the error between desired and selected color.
 	 Assume that, with a large scale (1024), we don't need to worry about
 	 image artifacts caused by error accumulation (the fact that the
 	 error terms might not sum to the error). */
-      e = (use_r - new_col[*new_data].red) * DITHER_SCALE;
-      if (e) {
-	r_err [x+d0] += (e * 7) / 16;
-	r_err1[x+d1] += (e * 3) / 16;
-	r_err1[x+d2] += (e * 5) / 16;
-	r_err1[x+d3] += e / 16;
-      }
-
-      e = (use_g - new_col[*new_data].green) * DITHER_SCALE;
-      if (e) {
-	g_err [x+d0] += (e * 7) / 16;
-	g_err1[x+d1] += (e * 3) / 16;
-	g_err1[x+d2] += (e * 5) / 16;
-	g_err1[x+d3] += e / 16;
-      }
-
-      e = (use_b - new_col[*new_data].blue) * DITHER_SCALE;
-      if (e) {
-	b_err [x+d0] += (e * 7) / 16;
-	b_err1[x+d1] += (e * 3) / 16;
-	b_err1[x+d2] += (e * 5) / 16;
-	b_err1[x+d3] += e / 16;
+      for (k = 0; k < 3; ++k) {
+          e = (use.a[k] - kd3->items[*new_data].a[k]) * DITHER_ITEM2ERR;
+          if (e) {
+              err [x+d0].a[k] += ((e * 7) & ~15) / 16;
+              err1[x+d1].a[k] += ((e * 3) & ~15) / 16;
+              err1[x+d2].a[k] += ((e * 5) & ~15) / 16;
+              err1[x+d3].a[k] += ( e      & ~15) / 16;
+          }
       }
 
      next:
@@ -835,21 +981,16 @@ colormap_image_floyd_steinberg(Gif_Image *gfi, uint8_t *all_new_data,
 
     /* change dithering directions */
     {
-      int32_t *temp;
-      temp = r_err; r_err = r_err1; r_err1 = temp;
-      temp = g_err; g_err = g_err1; g_err1 = temp;
-      temp = b_err; b_err = b_err1; b_err1 = temp;
+      color_erritem *temp = err1;
+      err1 = err;
+      err = temp;
       dither_direction = !dither_direction;
     }
   }
 
   /* delete temporary storage */
-  Gif_DeleteArray(r_err);
-  Gif_DeleteArray(g_err);
-  Gif_DeleteArray(b_err);
-  Gif_DeleteArray(r_err1);
-  Gif_DeleteArray(g_err1);
-  Gif_DeleteArray(b_err1);
+  Gif_DeleteArray(err);
+  Gif_DeleteArray(err1);
 }
 
 
@@ -857,7 +998,7 @@ colormap_image_floyd_steinberg(Gif_Image *gfi, uint8_t *all_new_data,
 static int
 try_assign_transparency(Gif_Image *gfi, Gif_Colormap *old_cm, uint8_t *new_data,
 			Gif_Colormap *new_cm, int *new_ncol,
-			uint32_t *histogram)
+			kd3_tree* kd3, uint32_t *histogram)
 {
   uint32_t min_used;
   int i, j;
@@ -902,7 +1043,7 @@ try_assign_transparency(Gif_Image *gfi, Gif_Colormap *old_cm, uint8_t *new_data,
       new_transparent = i;
       min_used = histogram[i];
     }
-  new_cm->col[new_transparent].haspixel = 255; /* mark it unusable */
+  kd3_disable(kd3, new_transparent); /* mark it as unusable */
   return 1;
 
  found:
@@ -921,10 +1062,10 @@ void
 colormap_stream(Gif_Stream *gfs, Gif_Colormap *new_cm,
 		colormap_image_func image_changer)
 {
-  color_hash_item **hash = new_color_hash();
+  kd3_tree kd3;
   int background_transparent = gfs->images[0]->transparent >= 0;
   Gif_Color *new_col = new_cm->col;
-  int new_ncol = new_cm->ncol;
+  int new_ncol = new_cm->ncol, new_gray;
   int imagei, j;
   int compress_new_cm = 1;
 
@@ -942,6 +1083,20 @@ colormap_stream(Gif_Stream *gfs, Gif_Colormap *new_cm,
   for (j = 0; j < 256; j++)
     new_col[j].pixel = 0;
 
+  /* initialize kd3 tree */
+  new_gray = 1;
+  for (j = 0; new_gray && j < new_cm->ncol; ++j)
+      if (new_col[j].gfc_red != new_col[j].gfc_green
+          || new_col[j].gfc_red != new_col[j].gfc_blue)
+          new_gray = 0;
+  if (new_gray)
+      kd3_init(&kd3, kd3_distance, kd3_luminance_transform);
+  else
+      kd3_init(&kd3, kd3_distance, NULL);
+  for (j = 0; j < new_cm->ncol; ++j)
+      kd3_add_8to15(&kd3, new_col[j].gfc_red, new_col[j].gfc_green, new_col[j].gfc_blue);
+  kd3_build(&kd3);
+
   for (imagei = 0; imagei < gfs->nimages; imagei++) {
     Gif_Image *gfi = gfs->images[imagei];
     Gif_Colormap *gfcm = gfi->local ? gfi->local : gfs->global;
@@ -957,11 +1112,13 @@ colormap_stream(Gif_Stream *gfs, Gif_Colormap *new_cm,
       if (only_compressed)
 	Gif_UncompressImage(gfi);
 
+      kd3_enable_all(&kd3);
       do {
-	for (j = 0; j < 256; j++) histogram[j] = 0;
-	image_changer(gfi, new_data, gfcm, new_cm, hash, histogram);
+	for (j = 0; j < 256; j++)
+            histogram[j] = 0;
+	image_changer(gfi, new_data, gfcm, &kd3, histogram);
       } while (try_assign_transparency(gfi, gfcm, new_data, new_cm, &new_ncol,
-				       histogram));
+				       &kd3, histogram));
 
       Gif_ReleaseUncompressedImage(gfi);
       /* version 1.28 bug fix: release any compressed version or it'll cause
@@ -1003,12 +1160,13 @@ colormap_stream(Gif_Stream *gfs, Gif_Colormap *new_cm,
   if (background_transparent)
     gfs->background = gfs->images[0]->transparent;
   else if (gfs->global && gfs->background < gfs->global->ncol) {
-    Gif_Color *c = &gfs->global->col[ gfs->background ];
-    gfs->background = hash_color(c->red, c->green, c->blue, hash, new_cm);
+    Gif_Color *c = &gfs->global->col[gfs->background];
+    gfs->background = kd3_closest_8to15(&kd3, c->gfc_red, c->gfc_green, c->gfc_blue);
     new_col[gfs->background].pixel++;
   }
 
   Gif_DeleteColormap(gfs->global);
+  kd3_cleanup(&kd3);
 
   /* We may have used only a subset of the colors in new_cm. We try to store
      only that subset, just as if we'd piped the output of 'gifsicle
@@ -1068,8 +1226,4 @@ colormap_stream(Gif_Stream *gfs, Gif_Colormap *new_cm,
       }
     }
   }
-
-  /* free storage */
-  free_all_color_hash_items();
-  Gif_DeleteArray(hash);
 }
